{%- if show_source and has_source and sourcename %}
  <div role="note" aria-label="source link">
    <h3>{{ _('This page') }}</h3>
    <ul class="this-page-menu">
      <li><a href="{{ pathto('bugs') }}">{% trans %}Report a bug{% endtrans %}</a></li>
      <li>
        <a href="https://github.com/python/cpython/blob/main/Doc/{{ sourcename|replace('.rst.txt', '.rst') }}?plain=1"
            rel="nofollow">{{ _('Show source') }}
        </a>
      </li>
      {% if language != "en" %}
      <li>
        <a href="https://github.com/python/python-docs-{{ language }}/blob/{{ version }}/{{ pagename }}.po?plain=1"
           rel="nofollow">{{ _('Show translation source') }}</a>
      </li>
      {% endif %}
    </ul>
  </div>
{%- endif %}
