{% extends "layout.html" %}
{%- block htmltitle -%}
<title>{{ shorttitle }}</title>
{%- endblock -%}
{%- block extrahead -%}
    {{ super() }}
    <meta property="og:title" content="Python {{ version }} documentation">
    <meta property="og:type" content="website">
    <meta property="og:url" content="https://docs.python.org/3/">
    <meta property="og:site_name" content="Python documentation">
    <meta property="og:description" content="The official Python documentation.">
    <meta property="og:image" content="https://docs.python.org/3/_static/og-image.png">
    <meta name="description" content="The official Python documentation.">
    <meta property="og:image:width" content="200">
    <meta property="og:image:height" content="200">
    <meta name="theme-color" content="#3776ab">
{%- endblock -%}
{% block body %}
  <h1>{{ docstitle|e }}</h1>
  <p>
  {% trans %}Welcome! This is the official documentation for Python {{ release }}.{% endtrans %}
  </p>
  <p><strong>{% trans %}Documentation sections:{% endtrans %}</strong></p>
  <table class="contentstable" align="center"><tr>
    <td width="50%">
      <p class="biglink"><a class="biglink" href="{{ pathto("whatsnew/" + version) }}">{% trans %}What's new in Python {{ version }}?{% endtrans %}</a><br>
         <span class="linkdescr"> {% trans whatsnew_index=pathto("whatsnew/index") %}Or <a href="{{ whatsnew_index }}">all "What's new" documents since Python 2.0</a>{% endtrans %}</span></p>
      <p class="biglink"><a class="biglink" href="{{ pathto("tutorial/index") }}">{% trans %}Tutorial{% endtrans %}</a><br>
         <span class="linkdescr">{% trans %}Start here: a tour of Python's syntax and features{% endtrans %}</span></p>
      <p class="biglink"><a class="biglink" href="{{ pathto("library/index") }}">{% trans %}Library reference{% endtrans %}</a><br>
         <span class="linkdescr">{% trans %}Standard library and builtins{% endtrans %}</span></p>
      <p class="biglink"><a class="biglink" href="{{ pathto("reference/index") }}">{% trans %}Language reference{% endtrans %}</a><br>
         <span class="linkdescr">{% trans %}Syntax and language elements{% endtrans %}</span></p>
      <p class="biglink"><a class="biglink" href="{{ pathto("using/index") }}">{% trans %}Python setup and usage{% endtrans %}</a><br>
         <span class="linkdescr">{% trans %}How to install, configure, and use Python{% endtrans %}</span></p>
      <p class="biglink"><a class="biglink" href="{{ pathto("howto/index") }}">{% trans %}Python HOWTOs{% endtrans %}</a><br>
         <span class="linkdescr">{% trans %}In-depth topic manuals{% endtrans %}</span></p>
    </td><td width="50%">
      <p class="biglink"><a class="biglink" href="{{ pathto("installing/index") }}">{% trans %}Installing Python modules{% endtrans %}</a><br>
         <span class="linkdescr">{% trans %}Third-party modules and PyPI.org{% endtrans %}</span></p>
      <p class="biglink"><a class="biglink" href="{{ pathto("distributing/index") }}">{% trans %}Distributing Python modules{% endtrans %}</a><br>
         <span class="linkdescr">{% trans %}Publishing modules for use by other people{% endtrans %}</span></p>
      <p class="biglink"><a class="biglink" href="{{ pathto("extending/index") }}">{% trans %}Extending and embedding{% endtrans %}</a><br>
         <span class="linkdescr">{% trans %}For C/C++ programmers{% endtrans %}</span></p>
      <p class="biglink"><a class="biglink" href="{{ pathto("c-api/index") }}">{% trans %}Python's C API{% endtrans %}</a><br>
         <span class="linkdescr">{% trans %}C API reference{% endtrans %}</span></p>
      <p class="biglink"><a class="biglink" href="{{ pathto("faq/index") }}">{% trans %}FAQs{% endtrans %}</a><br>
         <span class="linkdescr">{% trans %}Frequently asked questions (with answers!){% endtrans %}</span></p>
      <p class="biglink"><a class="biglink" href="{{ pathto("deprecations/index") }}">{% trans %}Deprecations{% endtrans %}</a><br>
         <span class="linkdescr">{% trans %}Deprecated functionality{% endtrans %}</span></p>
    </td></tr>
  </table>

  <p><strong>{% trans %}Indices, glossary, and search:{% endtrans %}</strong></p>
  <table class="contentstable" align="center"><tr>
    <td width="50%">
      <p class="biglink"><a class="biglink" href="{{ pathto("py-modindex") }}">{% trans %}Global module index{% endtrans %}</a><br>
         <span class="linkdescr">{% trans %}All modules and libraries{% endtrans %}</span></p>
      <p class="biglink"><a class="biglink" href="{{ pathto("genindex") }}">{% trans %}General index{% endtrans %}</a><br>
         <span class="linkdescr">{% trans %}All functions, classes, and terms{% endtrans %}</span></p>
      <p class="biglink"><a class="biglink" href="{{ pathto("glossary") }}">{% trans %}Glossary{% endtrans %}</a><br>
         <span class="linkdescr">{% trans %}Terms explained{% endtrans %}</span></p>
    </td><td width="50%">
      <p class="biglink"><a class="biglink" href="{{ pathto("search") }}">{% trans %}Search page{% endtrans %}</a><br>
         <span class="linkdescr">{% trans %}Search this documentation{% endtrans %}</span></p>
      <p class="biglink"><a class="biglink" href="{{ pathto("contents") }}">{% trans %}Complete table of contents{% endtrans %}</a><br>
         <span class="linkdescr">{% trans %}Lists all sections and subsections{% endtrans %}</span></p>
    </td></tr>
  </table>

  <p><strong>{% trans %}Project information:{% endtrans %}</strong></p>
  <table class="contentstable" align="center"><tr>
    <td width="50%">
      <p class="biglink"><a class="biglink" href="{{ pathto("bugs") }}">{% trans %}Reporting issues{% endtrans %}</a></p>
      <p class="biglink"><a class="biglink" href="https://devguide.python.org/documentation/help-documenting/">{% trans %}Contributing to docs{% endtrans %}</a></p>
      <p class="biglink"><a class="biglink" href="{{ pathto("download") }}">{% trans %}Download the documentation{% endtrans %}</a></p>
    </td><td width="50%">
      <p class="biglink"><a class="biglink" href="{{ pathto("license") }}">{% trans %}History and license of Python{% endtrans %}</a></p>
      <p class="biglink"><a class="biglink" href="{{ pathto("copyright") }}">{% trans %}Copyright{% endtrans %}</a></p>
      <p class="biglink"><a class="biglink" href="{{ pathto("about") }}">{% trans %}About the documentation{% endtrans %}</a></p>
    </td></tr>
  </table>
{% endblock %}
